from enum import Enum

import chess
import chess.pgn


def find_games(
    stream,
    /,
    query=None,
    time_controls=None,
    minimum_rating_bound=None,
    maximum_rating_bound=None,
    average_rating_bound=None,
    number_of_games=None,
):
    count = 0
    while True:
        game = chess.pgn.read_game(stream)
        if not game:
            break

        if count == number_of_games:
            break

        should_check_elo = (
            minimum_rating_bound or maximum_rating_bound or average_rating_bound
        )
        if should_check_elo:
            white_elo_not_specified = "WhiteElo" not in game.headers
            black_elo_not_specified = "BlackElo" not in game.headers
            if white_elo_not_specified or black_elo_not_specified:
                continue

            white_elo = int(game.headers["WhiteElo"])
            black_elo = int(game.headers["BlackElo"])
            minimum_elo = min(white_elo, black_elo)
            maximum_elo = max(white_elo, black_elo)
            average_elo = (white_elo + black_elo) / 2

        if minimum_rating_bound and minimum_elo < minimum_rating_bound:
            continue
        if maximum_rating_bound and maximum_elo > maximum_rating_bound:
            continue
        if average_rating_bound:
            if minimum_elo < (average_rating_bound - 50):
                continue
            if maximum_elo > (average_rating_bound + 50):
                continue

        if time_controls:
            no_time_control_info = "TimeControl" not in game.headers
            if no_time_control_info:
                continue

            time_control_info = game.headers["TimeControl"]

            if not any(
                time_control.matches(time_control_info)
                for time_control in time_controls
            ):
                continue

        if query and not query(game):
            continue

        count += 1
        yield game


class TimeControl(Enum):
    """
    Time controls are represented as bounds on the
    Estimated duration of a game in seconds
    """

    ULTRA_BULLET = (0, 29)
    BULLET = (30, 179)
    BLITZ = (180, 479)
    RAPID = (450, 1499)
    CLASSICAL = (1500, 21599)
    CORRESPONDENCE = (21600, 2147483647)

    def matches(self, other):
        if "+" not in other:
            return False
        clock, increment = other.split("+")
        lower_bound, upper_bound = self.value

        return lower_bound < int(clock) + 40 * int(increment) <= upper_bound



import importlib.util
import sys
from argparse import ArgumentParser

from .search import TimeControl, find_games

ERROR_COLOUR = "\033[91m"
RESET_COLOUR = "\033[0m"

def main():
    parser = ArgumentParser(
        prog="pgn-filter",
        description="A small program to query for games inside PGN documents",
        epilog="""
        Every month I look through some ten thousand games
        ~ Vladimir Kramnik
        """,
    )

    parser.add_argument("-f", "--file", help="The PGN file to search through")
    parser.add_argument("-i", "--stdin", action="store_true", help="Read from STDIN")
    parser.add_argument(
        "-q", "--query", help="The Python file containing the query to use"
    )
    parser.add_argument(
        "-n",
        "--number-of-games",
        metavar="number",
        type=int,
        help="The maximum number of games to return",
    )
    parser.add_argument(
        "-m",
        "--minimum-rating",
        metavar="rating",
        type=int,
        help="The minimum rating of games to consider",
    )
    parser.add_argument(
        "-M",
        "--maximum-rating",
        metavar="rating",
        type=int,
        help="The maximum rating of games to consider",
    )
    parser.add_argument(
        "-a", "--average-rating", metavar="rating", type=int,help="The rating range to consider"
    )
    parser.add_argument(
        "-F", "--fast", action="store_true", help="Only consider bullet games"
    )
    parser.add_argument(
        "-S", "--slow", action="store_true", help="Only consider blitz games and slower"
    )

    arguments = parser.parse_args()

    error = False
    if not arguments.file and not arguments.stdin:
        print(ERROR_COLOUR + "Error: You must specify an input source")
        error = True

    if arguments.file and arguments.stdin:
        print(
            ERROR_COLOUR
            + "Error: Cannot read from a file and while reading from sdtandard input",
        )
        error = True

    if arguments.fast and arguments.slow:
        print(ERROR_COLOUR + "Error: --fast and --slow cannot be used together")
        error = True

    print(RESET_COLOUR, end="")

    if error:
        parser.print_help()
        exit()

    stream = None
    if arguments.file:
        stream = open(arguments.file)

    if arguments.stdin:
        stream = sys.stdin

    query = None
    if arguments.query:
        specification = importlib.util.spec_from_file_location("query", arguments.query)
        module = importlib.util.module_from_spec(specification)
        specification.loader.exec_module(module)
        query = module.query

    time_controls = list(TimeControl)
    if arguments.fast:
        time_controls = [TimeControl.ULTRA_BULLET, TimeControl.BULLET]

    if arguments.slow:
        time_controls = [
            TimeControl.BLITZ,
            TimeControl.RAPID,
            TimeControl.CORRESPONDENCE,
        ]

    minimum_rating = arguments.minimum_rating
    maximum_rating = arguments.maximum_rating
    average_rating = arguments.average_rating
    games = find_games(
        stream,
        query=query,
        time_controls=time_controls,
        minimum_rating_bound=minimum_rating,
        maximum_rating_bound=maximum_rating,
        average_rating_bound=average_rating,
        number_of_games=arguments.number_of_games
    )

    has_tqdm = False
    try:
        from tqdm import tqdm
        has_tqdm = True
    except ImportError:
        pass

    should_display_progress_bar = has_tqdm and arguments.number_of_games
    if should_display_progress_bar:
        progress_bar = tqdm(total=arguments.number_of_games)

    for game in games:
        if should_display_progress_bar:
            progress_bar.update(1)
        print(game, end="\n\n")

    stream.close()


if __name__ == "__main__":
    main()
